/*
 *  UQLX
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/times.h>
#include <time.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h> 
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#ifdef VM_SCR
#include <sys/mman.h>
#endif
  

#include "QLtypes.h"
#include "QL68000.h"
#include "unix.h"

#include "uqlx_cfg.h"

#define STATIC static

#ifdef SH_MEM
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

static XShmSegmentInfo shminfo;
static shm_used;
#endif

XSetWindowAttributes attributes;

int depth,inside;
long black,white;
extern int scr_width,scr_height;

extern void QLButton(int,int);
extern void QLMovePointer(int,int);

#if 0
extern char *scrModTable;
extern int sct_size;
extern int pagesize;
#endif

extern int verbose;

#define E_MASK ( KeyPressMask | KeyReleaseMask | ButtonPressMask	\
		 | ButtonReleaseMask | PointerMotionMask | ColormapChangeMask |\
		 EnterWindowMask | LeaveWindowMask | ExposureMask | ResizeRedirectMask )

     extern void QL2Pixmap8 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap16 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap24 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap32 (w32 , char *, w32 ,w32 );

     extern void (*QL2Pixmap) (w32 , char *, w32 ,w32 );
     

static int shmmono=1;
static char win_name[128];
static Visual *visual = NULL;

static int screen, nplanes;
static int Max_x, Max_y;

GC gc;
static XGCValues gc_val;

Window imagewin;		/* window ID for image */
XImage *image = NULL;
static char *Image;
Cursor cursor;
Pixmap empty;
int planes,plane2;
int pbytes;


char *xi_buf;

#ifdef SH_MEM
static int shmct;
#endif

Display *display;
int kmin,kmax;
int shmflag = 1;
Colormap cmap=0;

void redraw_screen(int ,int ,int ,int );


static void destroy_image (void)
{
  if (image == NULL) return;
  XDestroyImage (image);
#ifdef SH_MEM
  if (shmflag && shm_used)
    {
      XShmDetach (display, &shminfo);
      shmdt (shminfo.shmaddr);
      shmctl (shminfo.shmid, IPC_RMID, 0);
      shm_used = 0;
    }
#endif
}

void x_screen_close (void)
{
  XAutoRepeatOn(display);
  XFlush(display);
  destroy_image();
}

static void set_window_name (char *name)
{
  XStoreName (display, imagewin, name);
}




static void create_image (int w, int h, int format, int frk)
{
  int planes = depth;
  
#ifdef SH_MEM
  if (shmflag)
    {
      image = XShmCreateImage (display, visual, planes, format,
			       NULL, &shminfo, w, h);
      shminfo.shmid = shmget (IPC_PRIVATE, image->bytes_per_line
			      * image->height * ((planes+7)/8), IPC_CREAT | 0777);
      xi_buf = shminfo.shmaddr = image->data = shmat (shminfo.shmid, 0, 0);
      shminfo.readOnly = False;
      XShmAttach (display, &shminfo);
      XSync(display, False);
      shmctl(shminfo.shmid, IPC_RMID, 0);
      shm_used = 1;
#if 0
      bzero (&shminfo, sizeof(shminfo));
      printf("Using MIT Shm extension::\n\t\t If you get trouble with X errors\n");
      printf("\t\ttry to compile wthout the -DSH_MEM flag\n\n");
#endif

#ifdef BIG_ENDIAN
      image -> bitmap_bit_order = MSBFirst;
      image -> byte_order = MSBFirst;
#else
      image -> bitmap_bit_order = LSBFirst;
      image -> byte_order = LSBFirst;
#endif

      XShmPutImage(display,imagewin,gc,image,0,0,0,0,10,10,False);
      XSync(display,False);
    }

  /* XShmPutImage might have failed, setting shmflag=0 ! */

  if (!shmflag)
#endif
    {
      /*      if (planes==8 || planes==16)
	xi_buf = malloc (w * h  *  ((planes+7)/8));
      else*/

      xi_buf= malloc(w*h*pbytes+256);
      
      image = XCreateImage (display, visual, planes, format,
			    0, (char *) xi_buf, w, h, 8, 0);
    }

#ifdef BIG_ENDIAN
  image -> bitmap_bit_order = MSBFirst;
  image -> byte_order = MSBFirst;
#else
  image -> bitmap_bit_order = LSBFirst;
  image -> byte_order = LSBFirst;
#endif
  /*Image = image->data;*/

}


/* Process all events that occured since we last came here...
 */

extern void xql_key(XEvent *,int);

void process_events (void)
{
  XEvent e;
  XExposeEvent *x;
  while (XCheckMaskEvent (display, E_MASK, &e))
    {
      switch (e.type)
	{
	case EnterNotify:
	  XAutoRepeatOff(display); /* QL has its autorepeat.. */
	  inside=1;
	  
	  break;
	case LeaveNotify:
	  XAutoRepeatOn(display);
	  inside=0;
	  
	  break;
	case ResizeRequest:
	  /* move way of making it not resizeable... */
	  XResizeWindow(display, imagewin, scr_width, scr_height);
	case KeyPress:
	  xql_key (&e, 1);
	  inside=1;
	  break;
	case KeyRelease:
	  xql_key (&e, 0);
	  break;
#ifdef MOUSE   /*mouse not yet supported */
	case ButtonPress:
	  QLButton(((XButtonEvent *)&e)->button,1);
	  inside=1;
	  break;
#endif
#ifdef MOUSE  /* no mouse yet */
	case ButtonRelease:
	  QLButton(((XButtonEvent *)&e)->button,0);
	  break;
#endif 

	case Expose:
	  x = (XExposeEvent *)&e;
	  redraw_screen(x->x,x->y,x->width,x->height);	  
	  break;

#ifdef MOUSE 
	case MotionNotify: 
	  if (e.xmotion.window == imagewin)
	    {
	      QLMovePointer(((XMotionEvent *)&e)->x,((XMotionEvent *)&e)->y);
	      /*fprintf (stderr, "Pointer: %d,%d\n", ((XMotionEvent *)&e)->x,((XMotionEvent *)&e)->y); */
	      inside=1;
	    }
	  break;
#endif 0
	}
    }
}

long colors[8];

void allocColors(Colormap map)
{
  static char *Cname[]={"Blue","Red","Magenta","Green","Cyan","Yellow","gray95" };
  static char *Gname[]={"gray15","gray30","gray45","gray60","gray75","gray90","gray95"};
  char **name;
  
  XColor exact_def;
  int default_depth=DefaultDepth(display,screen);
  int res,i=5;

  XVisualInfo visual_info;

  static char *visual_class[]={"StaticGray","GrayScale","StaticColor",
			       "PseudoColor","TrueColor","DirectColor"};

  planes=default_depth;
  plane2=planes/2;
  

  map=DefaultColormap(display,screen);
  XSetWindowColormap(display,imagewin,map);

  if (default_depth<4)  {
    printf("sorry, not much support fo monochrome display\n");
    colors[0]=colors[1]=black;
    colors[2]=colors[3]=white;
    colors[4]=colors[5]=black;
    colors[6]=colors[7]=white;
  }  

  while(!XMatchVisualInfo(display,screen,default_depth,i--,&visual_info));
  
  if (verbose) printf("using Visual %s\n",visual_class[i++]);
  

  if (i>1 && QMD.color) name=Cname;
  else name=Gname;


#define MAXCOLORS 7
 
 
  if (verbose) printf("Default Visual %s\n",visual_class[DefaultVisual(display,screen)->class]); 
  
   colors[0]=colors[1]=BlackPixel(display,screen);
  colors[7]=colors[6]=WhitePixel(display,screen);  

  for(i=0; i<MAXCOLORS; i++){
    res=XParseColor(display,map,name[i],&exact_def);
    /*printf("XParseColor returns: %d\n",res);*/
    res=XAllocColor(display,map,&exact_def);
    /*printf("XAllocColor returns: %d\n",res);*/
    colors[(i+1)]=exact_def.pixel;/*colors[2*(i+1)]=colors[2*(i+1)+1]=exact_def.pixel;*/
    /*printf("pixval : %d\n",exact_def.pixel);*/
  }

  /*  for(i=0; i<8; i++)printf("Color %d= %d\n",i,colors[i]);*/
}


int myXerr(Display* d, XErrorEvent* ev)
{
  if (shmflag) shmflag=0;
  else
    printf("X error : trying to ignore\n");
}


void x_screen_open (int frk)
{
  unsigned long valuemask;
  int i;
  char *ip;
  
#ifdef SH_MEM
  int shm_major, shm_minor;
  Bool shm_pixmaps;
#endif

  XSizeHints h;
  XColor ccls[2];


  scr_width = qlscreen.xres;
  scr_height = qlscreen.yres;


  /*strcpy (win_name, "QL");*/
  sprintf(win_name,"QL - %s, %dK",QMD.sysrom,RTOP/1024);

  if ((display = XOpenDisplay (NULL)) == NULL)
    {
      perror(NULL);
      fprintf (stderr, "Could not open display.\n");
      exit (1);
    }
 
  if (!frk)
    init_keymap();
  
  /* set X error handler */
  XSetErrorHandler(myXerr);
#ifdef SH_MEM 
  if (shmflag)
    {
      if (!XShmQueryVersion (display, &shm_major, &shm_minor, &shm_pixmaps))
	{
	  verbose && fprintf(stderr,
			     "MIT Shared Memory Extension not supported.\n");
	  shmflag = 0;
	}
      else
	{
	  int foo;
	  verbose && fprintf (stderr,
			      "Using MIT Shared Memory Extension %d.%d," \
			      " %s shared pixmaps.\n", shm_major, shm_minor,
			      (shm_pixmaps ? "with" : "without"));
	  foo = XShmPixmapFormat (display);
#if 0
	  verbose && fprintf (stderr,
			      "XShm Pixmap format: %s\n", ((foo == XYBitmap ? 
							    "XYBitmap" : (foo == XYPixmap ? "XYPixmap" : "ZPixmap"))));
#endif
	}
    }
#endif
	
  screen = XDefaultScreen (display);
  depth = DefaultDepth (display, screen);  

  visual = XDefaultVisual (display, screen);
  nplanes = XDisplayPlanes (display, screen);

  if (nplanes<=16) pbytes=(nplanes+7)/8;
  else pbytes=4;
  
#if 0
    if (shmflag)
    {
      char *c=malloc(64);
      t = XCreateImage (display, visual, 1, XYBitmap, 0, c, 32, 16, 8, 0);
      if (t->bitmap_bit_order == LSBFirst)
	{
	  fprintf (stderr, "NOTE: Because XShmPutImage apparently can't swap bits, shared memory support\nwas turned off for monochrome mode!\n");
	  shmmono=0;
	}
      XDestroyImage (t);
    }
#endif

  h.width = scr_width;
  h.height = scr_height;
  h.x = 500;
  h.y = 0;

  h.flags = PSize | PPosition;
  black=BlackPixel (display, screen);
  white = WhitePixel (display, screen);

  attributes.backing_store=Always;
  valuemask=CWBackingStore;
 
  imagewin = XCreateWindow (display, RootWindow (display, screen),
			    h.x, h.y, h.width, h.height, 2, depth, InputOutput, visual,
			    valuemask, &attributes);

  XSelectInput (display, imagewin, StructureNotifyMask|E_MASK);
  XSetStandardProperties (display, imagewin, win_name, win_name, None,
			  NULL, 0, &h);
  XMapWindow (display, imagewin);
  for (;;)
    {
      XEvent e;
      XNextEvent (display, &e);
      if (e.type == MapNotify && e.xmap.event == imagewin) break;
    }
  gc_val.background = black;
  gc_val.foreground = white;
  gc = XCreateGC (display, imagewin, GCForeground | GCBackground, &gc_val);
#ifdef SH_MEM
  shmct = XShmGetEventBase (display) + ShmCompletion;
#endif
  XSetForeground (display, gc, white);
  XSetBackground (display, gc, black);

  if(!frk)
    allocColors(cmap);

  XClearWindow (display, imagewin);

  XSelectInput (display, imagewin, E_MASK);

  if (!frk)
    create_image (scr_width,scr_height,ZPixmap,frk);

  switch(nplanes)
    {
    case 8: QL2Pixmap=QL2Pixmap8;
      break;
    case 16: QL2Pixmap=QL2Pixmap16;
      break;
    case 24: QL2Pixmap=QL2Pixmap24;
      break;
    case 32: QL2Pixmap=QL2Pixmap32;
      break;
    default: printf("sorry, unsupported screen depth %d\n",nplanes);
      QL2Pixmap=QL2Pixmap8;
      break;
    }
#if 0
  ip=xi_buf;
  for (i=0; i<scr_width*scr_height*((nplanes+7)/8); i+=(nplanes+7)/8) *ip++=colors[0];
#endif 0
#if 0
   XPutImage(display, imagewin, gc, image,
		 0, 0,        
		 0, 0,        
		 512, 256 ); 
#endif

  
  ccls[0].pixel=0;
  ccls[1].pixel=0;
  empty =  XCreatePixmapFromBitmapData (display, imagewin, calloc(32,1),
					16, 16,0,0,1);
  cursor = XCreatePixmapCursor (display,empty,empty, &ccls[0],&ccls[1], 0,0);

  /* Set Passive Grabs for every key that could be snatched away byt the WM */

#ifdef DO_GRAB
  XGrabKey(display,AnyKey,AnyModifier,imagewin,False,GrabModeAsync,GrabModeAsync);
#endif

}

void ResetMousePtr()
{
  XUndefineCursor(display,imagewin);
}
void InvisibleMousePtr()
{
  XDefineCursor(display,imagewin,cursor);
}


